/*
    Trabalho Pratico II - IPDI
    -------------------------
    Anderson de Rezende Rocha <anderson.rocha@ic.unicamp.br>
    Leyza E. Baldo Dorini <leyzad@yahoo.com.br>

    Deteccao de circulos em imagens atraves da transformada de HOUGH 
    utilizando a direo do gradiente. Baseado no algoritmo proposto
    por Kimme, Ballard e Slansky
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include "imagem.h"
#include "morfo.h"
#include "utils.h"
#include "tratadorErros.h"

// Valor de PI = Comprimento de um circulo dividido por seu raio
#define PI 3.141592654

// Rotina para achar circulos na imagem de entrada atraves da transformada de Hough
void detectarCirculos(int *imagemCinza, int *imagemBinaria, int *imagemSaida, int nl, int nc);

int main (int argc, char *argv[])
{
    int OK, nc, nl, mn, OK2, nc2, nl2, mn2;
    int *imagemCinza;
    int *imagemBinaria;
    int *imagemSaida;    
    
    
    printf("\nInicio\n");
    if (argc == 4)
    {
        OK = le_imagem_pgm (argv[1], &imagemCinza, &nl, &nc, &mn);
        OK2 = le_imagem_pgm (argv[2], &imagemBinaria, &nl2, &nc2, &mn2);

        info_imagem (argv[1], nl, nc, mn);
        info_imagem (argv[2], nl2, nc2, mn2);
        
        aloca_memo (&imagemSaida, nl, nc);
        
        printf("Indo para a funo detectar circulos!! \n");       
        
        detectarCirculos(imagemCinza, imagemBinaria, imagemSaida, nl, nc);

        if(imagemCinza != NULL)
        {
            grava_imagem_pgm (imagemSaida, argv[3], nl, nc, mn);
            printf("Imagem \"%s\" gravada com sucesso!\n", argv[3]);
        }

        desaloca_memo(&imagemSaida);
    }
    else
    {
        tratarSintaxe(6);
    }

    printf("\n");
    return(0);
}

void detectarCirculos(int *imagemCinza, int *imagemBinaria, int *imagemSaida, int nl, int nc)
{
    int TAM_EP = (nl > nc ? nl : nc);
    int EP[TAM_EP][TAM_EP];   
    int imagemC[nl][nc];    
    int imagemB[nl][nc];    
    int i, j, xc, yc, s, flag, k, raio;
    float aux1, aux2, dgrad, delta, k1;
    int xc_maior,  yc_maior; 
    int vet_aux [2];
    float limiar;
    float aux = 0.0;
    k1 = 0.0;
    s = 1; //auxiliar para aproximao do gradiente

 
    for(i = 0; i < nl; i++)
    {
        for(j = 0; j < nc; j++)
        {
            // Le cada pixel e passa para a matriz
            imagemC[i][j] = *(imagemCinza + i * nc + j);
        }
    } 
    
    for(i = 0; i < nl; i++)
    {
        for(j = 0; j < nc; j++)
        {
            // Le cada pixel e passa para a matriz
            imagemB[i][j] = *(imagemBinaria + i * nc + j);
        }
    }

    printf("Inicializando o espao de parmetros...\n");
    for(i = 0; i < TAM_EP; i++)
    {
        for(j = 0; j < TAM_EP; j++)
        {   
            EP[i][j] = 0;            
        }
    }

    raio = 32;    
    delta = 1/raio; //variao do ngulo
    limiar = (raio/2)*raio; //para crculos com contorno completo funciona bem
                            //para os outros,  preciso calibrar...  limiar = c * raio
    k = 0.8 * raio; //porque 0.8 corresponde a aproximadamente 45 graus
    
    for(i = 0; i < nl; i++)
    {
        for(j = 0; j < nc; j++)
        {
            if(imagemB[i][j] != 255){
                flag=0;                
                //calcular o gradiente no ponto corresponde na imagem em tons de cinza               
                aux1 = (imagemC[i+s][j] - imagemC[i-s][j])/(2*s);
                aux2 = (imagemC[i][j+s] - imagemC[i][j-s])/(2*s);
                // se ambos forem diferentes de zero, faz as contas normal
                if(aux1!=0 && aux2!=0)
                {
                    dgrad = atan(aux2/aux1);
                    flag=1;
                }
                /* 
                se aux 1 for zero,  uma indeterminao. No entanto, pode-se considerar
                como o limite  de g2(x) sobre zero, que  +infinito ou -infinito. Na
                definio do arco-tangente, a tangente quando tende a +infinito  pi/2 e
                quando tende a menos infinito  -pi/2. Como o programa est considerando 
                os dois sentidos, tanto faz escolher um ou outro
                */
                else if(aux1==0 && aux2!=0)
                {
                    dgrad = PI/2;
                    flag=1;
                }
                /*
                Zero dividido por algo  zero...
                */
                else if(aux1!=0 && aux2==0)
                {
                    dgrad = 0;
                    flag=1;
                }                                    
                
                if(flag==1)
                {                             
                   for(k1=(-k/2); k1<=(k/2); k1++)
                   {  
                       /* 
                       Descrevendo uma poro de uma circunferncia de centro
                       (i,j) com o raio definido. A poro est definida no 
                       intervalo [dgrad-0.4, dgrad+0.4]
                       */
                       xc = i + raio * cos(dgrad + k1*delta);                            
                       yc = j + raio * sin(dgrad + k1*delta);
                       if((xc >= 0 && xc <= TAM_EP) && (yc>=0 && yc <= TAM_EP))
                            EP[xc][yc] +=1;                   
                   }                        

                   for(k1=(-k/2); k1<=(k/2); k1++)
                   {  
                       /* 
                       Est rodando 180 graus, para considerar o sentido inverso.
                       Ver Figura 4 do trabalho                        
                       */
                       xc = i + raio * cos(dgrad + PI + k1*delta);                            
                       yc = j + raio * sin(dgrad + PI + k1*delta);
                       if((xc >= 0 && xc <= TAM_EP) && (yc>=0 && yc <= TAM_EP))
                            EP[xc][yc] +=1;                   
                   }                 
                }                 
            }
        }   
    }
        
    //plotando os crculos encontrados
    for(i = 0; i < TAM_EP; i++)
    {
        for(j = 0; j < TAM_EP; j++)
        {                
             if(EP[i][j] > limiar)
             {
                 xc_maior = i;
                 yc_maior = j;
                 aux = 0.0;

                 while(aux <= (2*PI))
                 {
                     vet_aux[0] = xc_maior  + raio*cos(aux);
                     vet_aux[1] = yc_maior  + raio*sin(aux);
                     aux = aux + 0.01;
 
                     if((vet_aux[0]>0)&&(vet_aux[1]>0))
                     {
                         if((vet_aux[0] * nc + vet_aux[1]) < (nl * nc))
                         {
                             *(imagemSaida + vet_aux[0] * nc + vet_aux[1]) = 255;
                         }
                     }
                 }
             }    
        }
    }
}

